use color_eyre::Report;
use tokio::{
    io::{AsyncBufReadExt, AsyncWriteExt, BufReader},
    net::TcpListener,
    sync::broadcast,
};
use log::debug;
use local_ip_address::local_ip;

#[tokio::main]
async fn main() -> Result<(), Report> {
    env_logger::init();
    // Setup the Environment
    setup()?;

    let ip_address = local_ip().unwrap();

    // Set up a TCP listener to listen for incoming tcp requests at the ip address that is passed from the command line
    // TODO: Custom ports
    let listener = TcpListener::bind(format!("{ip_address}:8080")).await.unwrap();
    debug!("{}", format!("Bounded to {ip_address} port 8080"));

    // Setting up broadcast channel: A channel that will accept messages and broadcast them to everyone connected to the TCP server
    // the channel accepts an i32 that is the maximum number of messages the channel can retain at any given time
    let (sender, _receiver) = broadcast::channel(100);

    loop {
        // prevent the compiler error when we try to access something in a loop when it was initialized outside of the loop
        let sender = sender.clone();

        let mut receiver = sender.subscribe();

        // accept the requests
        let (mut socket, addr) = listener.accept().await.unwrap();
        debug!("Accepted a websocket request from {}", addr);

        tokio::spawn(async move {
            // Splitting the TCP socket into read/write halves
            let (reader, mut writer) = socket.split();

            let mut buffer_reader = BufReader::new(reader);

            let mut line: String = String::new();

            writer.write_all(format!("Your address is {addr} \n").as_bytes()).await.unwrap();

            // second loop keeps receiving input from that one client going
            loop {
                tokio::select! {
                    // input half
                    result = buffer_reader.read_line(&mut line) => {
                        if result.unwrap() == 0{
                            break;
                        }

                        // send items to the broadcast channel
                        sender.send((format!("{addr} - {line}"), addr)).unwrap();
                        // clear the input buffer
                        line.clear();
                    }

                    // print half
                    result = receiver.recv() => {
                        let (msg, other_addr) = result.unwrap();
                        if addr != other_addr {
                            writer.write_all(msg.as_bytes()).await.unwrap();
                        }
                    }
                }
            }
        });
    }
}

fn setup() -> Result<(), Report> {
    debug!("Changing some environment variables!");

    if std::env::var("RUST_LIB_BACKTRACE").is_err() {
        std::env::set_var("RUST_LIB_BACKTRACE", "1");
        debug!("RUST_LIB_BACKTRACE is set. There will be pretty error messages !")
    }
    color_eyre::install()?;

    Ok(())
}
